##########################################################################
# SmartPrintNG - high-quality export of Plone content to
# PDF, RTF, ODT, WML and DOCX
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

from zope.interface import Interface

class IConverter(Interface):

    def convert(context, html, **kw):
        """ A converter takes a snippet of HTML code and returns
            a valid XHTML page containing the snippet of HTML 
            in *some* way.
        """


class IHTMLExtractor(Interface):
    """ Extract a HTML snippet from the current object """

    def getHTML(context=None):
        """ Returns a HTML snippet to be converted for the current
            object. The method *must* not returned a full HTML page.
            An optional context object will be passed if the interface
            is implemented by an utility.
        """





##########################################################################
# SmartPrintNG - high-quality export of Plone content to
# PDF, RTF, ODT, WML and DOCX
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################


from Products.CMFCore.DirectoryView import registerDirectory

registerDirectory('skins', globals())

def initialize(context):

    import converters



##########################################################################
# SmartPrintNG - high-quality export of Plone content to
# PDF, RTF, ODT, WML and DOCX
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os
import tempfile
import logging
import threading

from Globals import InitializeClass
from Products.Five import BrowserView
from ZPublisher.Iterators import filestream_iterator

from zope.interface import providedBy
from zope.component import createObject, getAdapters, getAdapter
from zope.component.interfaces import ComponentLookupError

try:
    from zope.contenttype import guess_content_type
except ImportError:
    from OFS.content_types import guess_content_type

from Products.SmartPrintNG.interfaces import IHTMLExtractor
from standard_extractor import standard_html_extractor
from zcml import registry

C_LOCK = threading.Lock()
conversions_in_progres = 0

LOG = logging.getLogger()

class SmartPrintView(BrowserView):

    def convert(self, 
                html=None, 
                options=(),
                format='pdf', 
                content_extractor='',
                print_options={},
                converter_name='zopyx.smartprintng.converters.default',
                xslfo_converter_name='zopyx.smartprintng.converters.xslfo',
                redirect=False,
                **kw):
        """ Convert a html snippet to format.
            'html' - a HTML snippet. If not available IHTMLExtractor is being used
                     for extracting HTML fro the current object.

            'options' - a converter specific datastructure passed directly to the converter

            'print_optins' - dict of options passed as cmdline parameters down to css2xslfo

            'format' - rtf|pdf|pdf|odt|docx|wml

            'converter_name' - name of a registered converter utility
        """

        if content_extractor:
            # ATT: error handling
            try:
                extractor = getAdapter(self.context, IHTMLExtractor, name=content_extractor)
                html = extractor.getHTML(self.context)
            except ComponentLookupError:
                html = standard_html_extractor(self.context)

        # create up converter utility
        converter = createObject(converter_name)

        # 'html' should utf-8 encoded (from encodeURIComponent)
        html = converter.convert(self.context, html, options, **kw)

        # convert it back to 'utf-8', if necessary
        if isinstance(html, unicode):
            html = html.encode('utf-8')

        # Store HTML as a temporary file 
        tmpf = tempfile.mktemp()
        open(tmpf, 'wb').write(html)

        # and start the XSL-FO converter
        xslfoConverter = createObject(xslfo_converter_name, tmpf, encoding='utf-8')

        ok = False
        try:
            output_filename = xslfoConverter.convert(format, **dict(print_options))
            ok = True
        except:
            LOG.error('Error during conversion', exc_info=True)
            ok = False

        os.unlink(tmpf)

        if redirect:
            if ok:
                return self.request.RESPONSE.redirect(self.context.absolute_url() + 
                                                     '/smartPrintDeliver?filename=%s' % output_filename)
            raise RuntimeError('Error converting file')

        else:
            if ok:
                return output_filename
            raise RuntimeError('Error converting file')


    def deliver(self, filename):
        """ stream generated file back to client """

        def normalize(s):
            s = s.replace(' ', '_')
            return s

        type, enc = guess_content_type(filename)
        base, ext = os.path.splitext(os.path.basename(filename))
        id = normalize(self.context.getId())

        R = self.context.request.RESPONSE
        R.setHeader('content-type', type)
        R.setHeader('content-length', os.stat(filename)[6])
        R.setHeader('content-disposition', 'attachment; filename=%s%s' % (id, ext))
        return filestream_iterator(filename, 'rb')


    def availableFormats(self):
        """ Return a list of all support formats """
        from zopyx.convert import availableFormats
        return availableFormats()


    def availableTemplates(self):
        """ Return a list of tuples (template_name, description) """
        lst = list()

        # all registered interfaces for SmartPrintNG resources
        all_ifaces = registry.keys()

        # get all interfaces from the current object
        for iface in providedBy(self.context).flattened():

            # check if interface is registered for SmartPrintNG
            if iface in all_ifaces:
                d = registry[iface]

                # add list of available templates to result list
                for id, template in registry[iface]['templates']:
                    lst.append((id, id))

        return lst

    def availableStylesheets(self):
        """ Return a list of tuples (stylesheet_name, description) """
        lst = list()

        # all registered interfaces for SmartPrintNG resources
        all_ifaces = registry.keys()

        # get all interfaces from the current object
        for iface in providedBy(self.context).flattened():

            # check if interface is registered for SmartPrintNG
            if iface in all_ifaces:
                d = registry[iface]

                # add list of available templates to result list
                for id, template in registry[iface]['stylesheets']:
                    lst.append((id, id))

        return lst

    def _getStylesheet(self, name, registry_key):
        """ Return a FO related stylesheet by name """

        # all registered interfaces for SmartPrintNG resources
        all_ifaces = registry.keys()

        # get all interfaces from the current object
        for iface in providedBy(self.context).flattened():

            # check if interface is registered for SmartPrintNG
            if iface in all_ifaces:
                d = registry[iface]

                # add list of available templates to result list
                for id, sheet in registry[iface][registry_key]:
                    if id == name:
                        return open(sheet).read()

        raise ValueError('No stylesheet %s found' % name)


    def getFOStylesheet(self, name):
        """ Return a FO related stylesheet """
        return self._getStylesheet(name, 'fo_stylesheets')


    def getStylesheet(self, name):
        """ Return a standard stylesheet """
        return self._getStylesheet(name, 'stylesheets')


    def availableOptions(self):
        """ Return manipulations dict """
        from converter.manipulation import available_manipulations
        return available_manipulations


    def availableContentExtractors(self):
        """ return all names of registered IHTMLExtractor adapters for 
            the current context object.
        """
        return [name for name, adapter in getAdapters((self.context,), IHTMLExtractor)]

InitializeClass(SmartPrintView)



##########################################################################
# SmartPrintNG - high-quality export of Plone content to
# PDF, RTF, ODT, WML and DOCX
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os            
import random
import tempfile

from zope.interface import implements, implementedBy
from zope.component.interfaces import IFactory
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile

from Products.SmartPrintNG.interfaces import IConverter
from Products.SmartPrintNG.zcml import registry

from manipulation import available_manipulations, handleImages

# map mainpulation label -> method
manipulations_d = {}
for d in available_manipulations:
    manipulations_d[d['label']] = d['method']


class Converter(object):

    implements(IConverter)

    def __init__(self):
        self.id = str(random.random())
        self.tempdir = os.path.join(tempfile.gettempdir(), 'smartprinting', self.id)
        if not os.path.exists(self.tempdir):
            os.makedirs(self.tempdir)


    def convert(self, context, html, options, **kw):

        """ 'options' is a record-style datastructure with basically
            to attributes: 'workchain' representing a list of 
            HTML manipulation methods represented by their names and
            'template' represents to name of a template to render
            the full HTML.
        """

        # pre-process HTML fragment by running through a pipeline 
        # of converters/manipulation methods
        workchain = getattr(options, 'workchain', ())
        for d in available_manipulations:
            if d['label'] in workchain:
                html = d['method'](html)

        # now pass the modified HTML fragment to the template
        # in order to render a proper HTML file

        d = {'title' : 'Standardconverter',
             'body' : unicode(html, 'utf-8'), 
             'stylesheet' : getattr(options, 'stylesheet', 'default.css'),
            }

        # Search for the corresponding template in out registry
        for iface, d2 in registry.items():

            # first check if the interface is provided by the current
            # context object
            if not iface.providedBy(context):
                continue

            for name, template_filename in d2['templates']:
                if name == options.template:
                    template = ViewPageTemplateFile(template_filename)
                    # we need to set the debug flag to make the code
                    # work on Plone 2.5 aka Zope 2.9
                    context.request.debug = False
                    html = template(context, **d)
                    break

        # Last, fix all image links and export all related images 
        # into the filesystem
        html = handleImages(context, html, self.tempdir)

        return html


class Factory:
    
    implements(IFactory)

    def __init__(self, klass):
        self._klass = klass

    def __call__(self):
        return self._klass()

    def getInterfaces(self):
        return implementedBy(self._klass)

Factory = Factory(Converter)



##########################################################################
# SmartPrintNG - high-quality export of Plone content to
# PDF, RTF, ODT, WML and DOCX
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

""" HTML manipulation methods """

import re
import random
from cStringIO import StringIO
from urlparse import urlparse
import urllib2

from PIL import Image
from BeautifulSoup import BeautifulSoup, NavigableString, Tag

from Products.CMFCore.utils import getToolByName 

def removeLinks(html):
    """ replace all links with a <span> tag and the anchor text """

    soup = BeautifulSoup(html)
    refs = soup.findAll('a')

    for x in refs:
        tag = Tag(soup,'span')
        tag.insert(0,x.renderContents().strip())
        soup.a.replaceWith(tag)
    return  str(soup)

def removeReviewHistory(html):

    soup = BeautifulSoup(html)
    for x in soup.findAll(id='review-history'):
        x.extract()
    return  str(soup)


def removeImages(html):
    """ Remove all images """

    soup = BeautifulSoup(html)
    images = soup.findAll('img')
    [img.extract() for img in images]
    return  str(soup)


def checkHref(href):
    """ Return False for mailto|javascript or internal
        anchors or views.
    """

    if 'mailto:' in href or \
       'javascript:' in href or \
       href.startswith('#') or \
       href.startswith('@@'):
           return False
    return True
       

def getLinksInHtml(html):
    """ return all links inside a HTML fragment """

    soup = BeautifulSoup(html)
    hrefs = []
    for anchor in soup.findAll('a'):

        # we need to capture internal anchors
        try:
            href = anchor['href']
        except:
            continue 

        if href in hrefs or not checkHref(href):
           continue

        hrefs.append(str(href))
    return hrefs


def _pcdataFromNode(s, lst=[]):
    """ recursive pcdata collector """

    if s.string is not None:
        lst.append(s.string)
    else:
        for n in s.contents:
            _pcdataFromNode(n, lst)
    return ' '.join(lst)


def enumerateLinksInHtml(html, links):

    soup = BeautifulSoup(html)

    count = 1
    links = []
    for anchor in soup.findAll('a'):

        # capture internal anchors
        try:
            href = anchor['href']
        except KeyError:
            continue

        if not checkHref(href):
            continue

        anchor['class'] = 'enumerated-link'
        s = _pcdataFromNode(anchor) + ' [%d]' % count
        anchor.contents = [NavigableString(s)]
        links.append(anchor['href'])
        count += 1
        
    return str(soup)
    

def addLinkList(html):

    links = getLinksInHtml(html)
    html = enumerateLinksInHtml(html, links)

    if links:                      
        pat = '<li>[%d] %s</li>'
        link_string = '\n'.join([pat % (i+1, link) for i, link in enumerate(links)])
        linkLst = '\n<div class="enumerated-links"><ol>%s</ol</div>' % link_string
        html = html + linkLst

    return html


def handleImages(context, html, tempdir):

    mt = getToolByName(context, 'portal_membership')
    http_host = context.REQUEST.HTTP_HOST

    def fixImages(img):

        src = img['src'].encode('ascii')

        # we convert always to PNG in order to support *all*
        # external converters because they support different image types
        new_img = '%s/%s.png' % (tempdir, random.random())

        tp = urlparse(src)
        img_data = ''             

        same_host = http_host == tp[1]
        if tp[0] in ('http', 'https', 'ftp') and not same_host:
            # appears like a remote image, download it
            url = src
            img_data = urllib2.urlopen(url).read()
        else:
            # likely a local image, try to traverse to get hold
            # of the image data directly
            img_path = src
            if img_path.startswith('./'): # FCKeditor 
                img_path = img_path[2:]

            img_obj = None
            for path in (img_path, tp[2]):        
                img_obj = context.unrestrictedTraverse(path, None)
                if img_obj:
                    if mt.checkPermission('View', img_obj):
                        try:
                            img_data = str(img_obj.data)
                            break
                        except AttributeError:
                            try:
                                img_data = str(img_obj._data) # FSImage
                            except AttributeError:
                                break

        if img_data:        
            pil_img= Image.open(StringIO(img_data))
            pil_img.save(new_img, 'PNG')
            img['src'] = new_img
        else:
            img.extract()

    soup = BeautifulSoup(html)
    for img in soup.findAll('img'):
        fixImages(img)
    html = str(soup)
    return html


def breakIntoPages(html, seperator='(h1|h2)'):

    breaker = re.compile('<%s' % seperator, re.I|re.M|re.S)

    div_start = '<div class="chapter sp-page">'
    div_start2 = '<div class="chapter">'
    div_end = '</div>'

    positions = []
    for mo in breaker.finditer(html):
        positions.append(mo.start())
    positions.append(len(html))

    parts = []
    len_positions = len(positions) - 1
    for i in range(len_positions):
        start = positions[i]
        end = positions[i+1]

        if i == len_positions - 1:
            parts.append(div_start2 + html[start: end].strip() + div_end)
        else:
            parts.append(div_start + html[start: end].strip() + div_end)

    return '\n'.join(parts)



available_manipulations = (
    {'label' : 'removeImages', 'method' : removeImages, 'description': 'Remove images', 'state' : False},
    {'label' : 'removeLinks', 'method' : removeLinks, 'description': 'Remove links', 'state' : False},
    {'label' : 'addLinkList', 'method' : addLinkList, 'description': 'Add link list', 'state' : False},
    {'label' : 'breakIntoPages', 'method' : breakIntoPages, 'description': 'Break into pages', 'state' : False},
    {'label' : 'removeReviewHistory', 'method' : removeReviewHistory, 'description': 'Remove review history', 'state' : True},
)



##########################################################################
# SmartPrintNG
# (C) 2007, 2008, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os
from setuptools import setup, find_packages


CLASSIFIERS = [
    'Programming Language :: Python',
    'Framework :: Zope2',
    'Framework :: Plone',
]

version_file = os.path.join('Products', 'SmartPrintNG', 'version.txt')
version = open(version_file).read().strip()

readme_file= os.path.join('Products', 'SmartPrintNG', 'README.txt')
desc = open(readme_file).read().strip()
changes_file = os.path.join('Products', 'SmartPrintNG', 'CHANGES.txt')
changes = open(changes_file).read().strip()

long_description = desc + '\n\nCHANGES\n=======\n\n' +  changes 

setup(name='Products.SmartPrintNG',
      version=version,
      license='LGPL 3',
      author='Andreas Jung',
      author_email='info@zopyx.com',
      maintainer='Andreas Jung',
      maintainer_email='info@zopyx.com',
      classifiers=CLASSIFIERS,
      keywords='PDF  RTF  ODT  DOCX  WML conversion Plone Zope Python', 
      url='http://opensource.zopyx.com/projects/SmartPrintNG',
      description='Conversion of Plone content to PDF, RTS, ODT, DOCX and WML',
      long_description=long_description,
      packages=['Products', 'Products.SmartPrintNG'],
      include_package_data = True,
      zip_safe=False,
      install_requires=['setuptools', 'zopyx.convert>=1.1.7'],
      namespace_packages=['Products'],

      )



##########################################################################
# SmartPrintNG - high-quality export of Plone content to
# PDF, RTF, ODT, WML and DOCX
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

"""
Multipurpose HTML extractor for Plone content
"""

from BeautifulSoup import BeautifulSoup


def standard_html_extractor(obj):
    """ Extract HTML from an object by rendering it. """

    html = obj.view()
    soup = BeautifulSoup(html)
    body = str(soup.findAll('div', {'class' : 'documentContent'})[0])
    return body
